# Import packages
from io import StringIO
import pandas as pd
import numpy as np
import string
from unidecode import unidecode
import warnings
import time
import s3fs
import boto3
import jaro  # pip install jaro-winkler

# ------------------------------------
# -- Define cleaning function
# ------------------------------------


def clean_entity_name(original_entity: str, remove_stop_words: bool = True):

    clean_entity_name = str(original_entity)
    clean_entity_name = clean_entity_name.lower()
    clean_entity_name = clean_entity_name.split(" (")[0]
    translator = str.maketrans("", "", string.punctuation)
    clean_entity_name = clean_entity_name.translate(translator)
    clean_entity_name = unidecode(clean_entity_name)

    if remove_stop_words == True:
        # Remove stop words
        stop_words = [
            "limited",
            "ltd",
            "ag",
            "eg",
            "spa",
            "the",
            "sa",
            "ab",
            "llc",
            "llp",
            "plc",
            "inc",
            "ltd",
            "aktiengesellschaft",
            "nv",
            "a/s",
            "oc",
        ]
        clean_entity_name = " ".join(
            word for word in clean_entity_name.split() if word.lower() not in stop_words
        )

    return clean_entity_name


# ------------------------------------
# -- Define threshold function
# ------------------------------------


def dynamic_threshold(matching_scores: list, lower_threshold: float) -> float:

    # Sort the pairs based on the scores in descending order
    sorted_scores = sorted(matching_scores, reverse=True)
    # Remove scores below the lower threshold set in the main function
    sorted_scores = [x for x in sorted_scores if x >= lower_threshold]
    # Compute the ratio between any two consecutive pairs of scores
    score_ratios = [
        sorted_scores[i] / sorted_scores[i + 1] for i in range(len(sorted_scores) - 1)
    ]

    return sorted_scores[(score_ratios.index(max(score_ratios)))]


# ------------------------------------
# -- Define manual input of
# -- threshold function
# ------------------------------------


def threshold_check(
    matched_scores: list, matched_entities: list, current_threshold: float
) -> float:

    # Zip the two lists together
    combined_list = list(zip(matched_scores, matched_entities))
    # Sort the pairs based on the scores in descending order
    sorted_pairs = sorted(combined_list, reverse=True)
    # Separate the sorted pairs back into two lists
    sorted_scores, sorted_matches = zip(*sorted_pairs)

    # Keep only those entities above the threshold
    new_threshold = current_threshold
    while new_threshold != "":
        new_threshold = float(new_threshold)
        scores_to_keep = [x for x in sorted_scores if x >= new_threshold]
        matches_to_keep = sorted_matches[: len(scores_to_keep) + 1]
        first_three_discared_matches = sorted_matches[
            len(scores_to_keep) + 1 : len(scores_to_keep) + 4
        ]
        # matched_entities = [entity for entity, score in zip(matched_entities, matched_scores) if score > fixed_value]
        new_threshold = input(
            f"With a threshold of {calculated_threshold}, \
			the last three retained matches are {matches_to_keep[-3:]}, \
			while the first three discared ones are {first_three_discared_matches}. \
			Please input the new desired threshold (leave blank if current one is acceptable)"
        )

    return new_threshold


# ------------------------------------
# -- Define fuzzy matching function
# ------------------------------------


def run_fuzzy_match(
    entity_to_match: str,
    search_list: list,
    clean_names: bool = True,
    multiple_matches: bool = False,
    set_threshold_dynamically: bool = False,
    lower_threshold: float = 0.8,
) -> list:

    if clean_names == True:  # perform round of cleaning of names
        entity_to_match_clean = clean_entity_name(entity_to_match)
        search_list_clean = [clean_entity_name(x) for x in search_list]
    else:
        entity_to_match_clean = entity_to_match
        search_list_clean = search_list

    # Initialize empty lists
    matched_entities = []
    matched_scores = []

    # Compute a Jaro-Winkler score for each combination of entities
    for i, s in enumerate(search_list_clean):
        score = jaro.jaro_winkler_metric(entity_to_match_clean, s)
        matched_entities.append(search_list[i])
        matched_scores.append(score)

    # Extract matches of interest on the basis of the various parameters
    if multiple_matches == False:
        # Retain highest score only
        max_score = max(matched_scores)
        if max_score >= lower_threshold:
            matched_entities = matched_entities[
                matched_scores.index(max(matched_scores))
            ]
            matched_scores = max_score
        else:
            matched_scores = []
            matched_entities = []

    elif (multiple_matches == True) & (set_threshold_dynamically == False):
        matched_entities = [
            entity
            for entity, score in zip(matched_entities, matched_scores)
            if score >= lower_threshold
        ]
        matched_scores = [x for x in matched_scores if x >= lower_threshold]

    elif (multiple_matches == True) & (set_threshold_dynamically == True):
        # Calculate the dynamic threshold
        calculated_threshold = float(dynamic_threshold(matched_scores, lower_threshold))
        # Zip the two lists together
        combined_list = list(zip(matched_scores, matched_entities))
        # Sort the pairs based on the scores in descending order
        sorted_pairs = sorted(combined_list, reverse=True)
        # Separate the sorted pairs back into two lists
        sorted_scores, sorted_matches = zip(*sorted_pairs)
        # Retain all scores above or equal to the calculated threshold
        matched_scores = [x for x in sorted_scores if x >= calculated_threshold]
        matched_entities = sorted_matches[: len(matched_scores)]

    return matched_scores, matched_entities


# ------------------------------------
# -- Define master function
# ------------------------------------


def fuzzy_match_cpi(
    dataframe_to_match: pd.DataFrame,
    entity_column: str,
    search_list: list,
    clean_names: bool = True,
    multiple_matches: bool = False,
    set_threshold_dynamically: bool = False,
    lower_threshold: float = 0.8,
) -> pd.DataFrame:

    # Apply function
    results = dataframe_to_match[entity_column].apply(
        run_fuzzy_match,
        args=(
            search_list,
            clean_names,
            multiple_matches,
            set_threshold_dynamically,
            lower_threshold,
        ),
    )

    # Create two new columns to store the results
    dataframe_to_match["Matched score"] = results.apply(lambda x: x[0])
    dataframe_to_match["Matched entity"] = results.apply(lambda x: x[1])

    return dataframe_to_match


# Test
# importpath = '/Users/nikitamarini/Desktop/CPI'
# df = pd.read_csv(f'{importpath}/gfanz_entities.csv')
# search_list_test = ['Ageas', 'ABN AMRO', 'Intesa San Paolo']

# df = run_fuzzy_match(df, 'Entity', search_list_test)
# print(df[df['Matched entity'].apply(len) > 0])



import pandas as pd
import pandera as pa
from cpi_tools import aws_tools
import boto3
import s3fs

def get_valid_ref_values(ref_file, ref_column) -> list[str]:
    """ Retrieve valid GLCF column values from ref table"""
    ref_bucket = 'cpi-reference-data'
    ref_file_path = 'cpi/processed'
    ref_table = aws_tools.read_from_s3(ref_bucket, ref_file_path, ref_file)
    return list(ref_table[ref_column].unique())


def validate_glcf_data(df: pd.DataFrame, min_date: int = 2019, max_date:int = 2020) -> None:

    '''
    Input:

    df: pandas DataFrame containing GLCF data to be validated
    min_date (optional): integer representing the minimum year for valid data (default value is 2019)
    max_date (optional): integer representing the maximum year for valid data (default value is 2020)

    Returns: None

    This function validates a pandas DataFrame containing GLCF data, ensuring that it conforms to a specified schema. 
    The schema includes checks for valid values in various columns, such as sector, sub_sector, country, region, and institution type. 
    The function uses the validate method from the pandas_schema library to validate the DataFrame against the schema, 
    and raises an exception if the validation fails. If the validation is successful, the function prints a message 
    indicating that the DataFrame is GLCF compatible.
        
    '''

    # Get or define valid column values
    valid_sector_values = get_valid_ref_values('ref_project_type', 'sector')
    valid_sub_sector_values = get_valid_ref_values('ref_project_type', 'sub_sector')
    valid_solution_values = get_valid_ref_values('ref_project_type', 'solution')
    valid_country_values = get_valid_ref_values('ref_geog_list_cpi', 'country_cpi')
    valid_region_values = get_valid_ref_values('ref_geog_list_cpi', 'region_cpi')
    
    valid_institution_type_layer2_values =  ["Bilateral DFI", "Commercial FI", 'Corporation', 'Export Credit Agency (ECA)',
                            'Funds','Government', 'Household/Individual', 'Institutional Investors', 
                            'Multilateral Climate Funds','Multilateral DFI','National DFI', 'Public Fund',
                            'SOE', 'State-owned FI', 'Unknown']
    
    valid_institution_type_layer1_values = ['Public', 'Private']

    valid_domestic_international_values = ['Domestic', 'International']

    valid_use_values = ['Mitigation', 'Adaptation', 'Multiple Objectives']


    # Define GLCF Schema
    schema = pa.DataFrameSchema({
        "id_glcf": pa.Column(str),
        "data_source_id": pa.Column(str, required=False),
        "data_source": pa.Column(str),
        "year": pa.Column(int, checks=pa.Check.in_range(min_date, max_date)),
        "project_name": pa.Column(str),
        "project_description": pa.Column(str, nullable=True),
        "country_origin_cpi": pa.Column(str, pa.Check.isin(valid_country_values)),
        "region_origin_cpi": pa.Column(str, pa.Check.isin(valid_region_values)),
        "oecd_origin_cpi": pa.Column(str),
        "country_destination_cpi": pa.Column(str, pa.Check.isin(valid_country_values)),
        "region_destination_cpi": pa.Column(str, pa.Check.isin(valid_region_values)),
        "oecd_destination_cpi": pa.Column(str),
        "institution_cpi": pa.Column(str),
        "institution_type_layer1": pa.Column(str, pa.Check.isin(valid_institution_type_layer1_values)),
        "institution_type_layer2": pa.Column(str, pa.Check.isin(valid_institution_type_layer2_values)),
        "recipient_type_layer1": pa.Column(str),
        "domestic_international": pa.Column(str, pa.Check.isin(valid_domestic_international_values)),
        "value_USDm": pa.Column(float, pa.Check.ge(0)), #Greater than or equal to 0
        "instrument_cpi": pa.Column(str),
        "sector_original": pa.Column(str),
        "sector_key_cpi": pa.Column(str),
        "sector_cpi": pa.Column(str, checks=pa.Check.isin(valid_sector_values)),
        "sub_sector_cpi": pa.Column(str, pa.Check.isin(valid_sub_sector_values)),
        "solution_cpi": pa.Column(str, pa.Check.isin(valid_solution_values)),
        "use_cpi": pa.Column(str, pa.Check.isin(valid_use_values)),
        "gender_cpi": pa.Column(str),
        "exclude_include": pa.Column(str)
    })

    # Attempt Validation
    try:
        schema.validate(df, lazy=True)
        print('Dataframe GLCF Compatible')
    except pa.errors.SchemaErrors as e:
        #Display Errors upon validation failure
        display(e.failure_cases)



    
    
    
    
    
    